@@ -1,4 +1,4 @@
-/* $Id: mkzftree.c,v 1.11 2001/07/27 19:36:43 hpa Exp $ */
+/* $Id: mkzftree.c,v 1.16 2002/11/07 04:10:06 hpa Exp $ */
 /* ----------------------------------------------------------------------- *
  *   
  *   Copyright 2001 H. Peter Anvin - All Rights Reserved
@@ -77,28 +77,33 @@
 #endif
 
 /* Command line options */
-int force = 0;			/* Always compress */
-int level = 9;			/* Compression level */
-int parallel = 0;		/* Parallelism (0 = strictly serial) */
-int onefs = 0;			/* One filesystem only */
-int onedir = 0;			/* One directory only */
-int do_mkdir = 1;		/* Create stub directories */
-enum verbosity verbosity = default_verbosity;	/* Default verbosity */
-munger_func munger = block_compress_file; /* Default action */
+struct cmdline_options opt = {
+  0,				/* Force compression */
+  9,				/* Compression level */
+  0,				/* Parallelism (0 = strictly serial) */
+  0,				/* One filesystem only */
+  0,				/* One directory only */
+  1,				/* Create stub directories */
+  0,				/* Root may be a file */
+  0,				/* Be paranoid about metadata */
+  default_verbosity,		/* Default verbosity */
+  block_compress_file		/* Default transformation function */
+};
 
 /* Program name */
 const char *program;
 
 /* Long options */
-#define OPTSTRING "fl:up:xXlLFvqV:hw"
+#define OPTSTRING "fz:up:xXC:lLFvqV:hw"
 #ifdef HAVE_GETOPT_LONG
 const struct option long_options[] = {
   { "force",	             0,  0,  'f' },
-  { "level",                 1,  0,  'l' },
+  { "level",                 1,  0,  'z' },
   { "uncompress",            0,  0,  'u' },
   { "parallelism",           1,  0,  'p' },
   { "one-filesystem",        0,  0,  'x' },
   { "strict-one-filesystem", 0,  0,  'X' },
+  { "crib-tree",             1,  0,  'C' },
   { "local",                 0,  0,  'l' },
   { "strict-local",          0,  0,  'L' },
   { "file",                  0,  0,  'F' },
@@ -126,9 +131,11 @@ static void usage(enum verbosity level, int err)
 	  LO("  --parallelism #        ")"  -p #  Process up to # files in parallel\n"
 	  LO("  --one-filesystem       ")"  -x    Do not cross filesystem boundaries\n"
 	  LO("  --strict-one-filesystem")"  -X    Same as -x, but don't create stubs dirs\n"
+	  LO("  --crib-tree            ")"  -C    Steal \"crib\" files from an old tree\n"
 	  LO("  --local                ")"  -l    Do not recurse into subdirectoires\n"
 	  LO("  --strict-local         ")"  -L    Same as -l, but don't create stubs dirs\n"
 	  LO("  --file                 ")"  -F    Operate possibly on a single file\n"
+	  LO("  --sloppy               ")"  -s    Don't abort if metadata cannot be set\n"
 	  LO("  --verbose              ")"  -v    Increase message verbosity\n"
 	  LO("  --verbosity #          ")"  -V #  Set message verbosity to # (default = %d)\n"
 	  LO("  --quiet                ")"  -q    No messages, not even errors (-V 0)\n"
@@ -153,62 +160,62 @@ static int opt_atoi(const char *str)
 
 int main(int argc, char *argv[])
 {
-  const char *in, *out;
+  const char *in, *out, *crib = NULL;
   struct stat st;
   struct utimbuf ut;
-  int opt, err;
-  int file_root = 0;
+  int optch, err;
 
   program = argv[0];
 
-  while ( (opt = getopt_long(argc, argv, OPTSTRING, long_options, NULL))
+  while ( (optch = getopt_long(argc, argv, OPTSTRING, long_options, NULL))
 	  != EOF ) {
-    switch(opt) {
+    switch(optch) {
     case 'f':
-      force = 1;		/* Always compress */
+      opt.force = 1;		/* Always compress */
       break;
     case 'z':
-      if ( optarg[0] < '0' || optarg[0] > '9' || optarg[1] ) {
-	message(vl_error, "%s: invalid compression level: %s\n",
+      opt.level = opt_atoi(optarg);
+      if ( opt.level < 1 || opt.level > 9 ) {
+	message(vl_error, "%s: invalid compression level: %d\n",
 		program, optarg);
 	exit(EX_USAGE);
-      } else {
-	level = optarg[0] - '0';
       }
       break;
     case 'v':
-      verbosity++;
+      opt.verbosity++;
       break;
     case 'V':
-      verbosity = opt_atoi(optarg);
+      opt.verbosity = opt_atoi(optarg);
       break;
     case 'q':
-      verbosity = vl_quiet;
+      opt.verbosity = vl_quiet;
       break;
     case 'u':
-      munger = block_uncompress_file;
+      opt.munger = block_uncompress_file;
+      break;
+    case 'C':
+      crib = optarg;
       break;
     case 'p':
-      parallel = opt_atoi(optarg);
+      opt.parallel = opt_atoi(optarg);
       break;
     case 'x':
-      onefs = 1;
-      do_mkdir = 1;
+      opt.onefs = 1;  opt.do_mkdir = 1;
       break;
     case 'l':
-      onedir = 1;
-      do_mkdir = 1;
+      opt.onedir = 1; opt.do_mkdir = 1;
       break;
     case 'X':
-      onefs = 1;
-      do_mkdir = 0;
+      opt.onefs = 1;  opt.do_mkdir = 0;
       break;
     case 'L':
-      onedir = 1;
-      do_mkdir = 0;
+      opt.onedir = 1; opt.do_mkdir = 0;
       break;
     case 'F':
-      file_root = 1;
+      opt.file_root = 1;
+      break;
+    case 's':
+      opt.sloppy = 1;
       break;
     case 'h':
       usage(vl_quiet, 0);
@@ -223,20 +230,20 @@ int main(int argc, char *argv[])
   }
 
   if ( (argc-optind) != 2 )
-    usage(vl_error, 1);
+    usage(vl_error, EX_USAGE);
 
   in  = argv[optind];		/* Input tree */
   out = argv[optind+1];		/* Output tree */
 
   umask(077);
 
-  if ( file_root ) {
+  if ( opt.file_root ) {
     if ( lstat(in, &st) ) {
       message(vl_error, "%s: %s: %s\n", program, in, strerror(errno));
       exit(EX_NOINPUT);
     }
 
-    err = munge_entry(in, out, NULL);
+    err = munge_entry(in, out, crib, NULL);
   } else {
     /* Special case: we use stat() for the root, not lstat() */
     if ( stat(in, &st) ) {
@@ -248,7 +255,7 @@ int main(int argc, char *argv[])
       exit(EX_DATAERR);
     }
     
-    err = munge_tree(in, out);
+    err = munge_tree(in, out, crib);
   }    
 
   wait_for_all_workers();
@@ -256,13 +263,22 @@ int main(int argc, char *argv[])
   if ( err )
     exit(err);
 
-  if ( !file_root ) {
-    chown(out, st.st_uid, st.st_gid);
-    chmod(out, st.st_mode);
+  if ( !opt.file_root ) {
+    if ( chown(out, st.st_uid, st.st_gid) && !opt.sloppy ) {
+      message(vl_error, "%s: %s: %s", program, out, strerror(errno));
+      err = EX_CANTCREAT;
+    }
+    if ( chmod(out, st.st_mode) && !opt.sloppy && !err ) {
+      message(vl_error, "%s: %s: %s", program, out, strerror(errno));
+      err = EX_CANTCREAT;
+    }
     ut.actime  = st.st_atime;
     ut.modtime = st.st_mtime;
-    utime(out, &ut);
+    if ( utime(out, &ut) && !opt.sloppy && !err ) {
+      message(vl_error, "%s: %s: %s", program, out, strerror(errno));
+      err = EX_CANTCREAT;
+    }
   }
 
-  return 0;
+  return err;
 }
